/* SPDX-License-Identifier: LGPL-2.1-or-later */
#popover-machines-connection-selector-popover-body {
    padding-inline-end: 0 !important;

    ul {
        padding-inline-start: 0 !important;
    }
}
